@@ -41,13 +41,13 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         None.
 
         '''
-        print("ELMEnsemble __init__")
+        # print("ELMEnsemble __init__")
         self.n_estimators = n_estimators
         self.n_neurons = n_neurons
         self.activation = activation
         self.weight_distr = weight_distr
         self.weight_scl = weight_scl              
-        self.n_jobs = n_jobs                        ## a paraleliser
+        self.n_jobs = n_jobs                        ## to parallelize
         self.random_state = random_state
         
     def fit(self, X, y):
@@ -67,8 +67,14 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         -------
         Self.
         
+        Reference
+        ---------
+        G.-B. Huang, Q.-Y. Zhu, C.-K. Siew, 
+        Extreme learning machine: theory and applications, 
+        Neurocomputing 70 (1-3) (2006) 489–501.  
+        
         '''    
-        print("ELMEnsemble fit")                
+        # print("ELMEnsemble fit")                
         self.check = X.shape
         X, y = check_X_y(X, y)
         
@@ -112,7 +118,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Predicted output
 
         ''' 
-        print("ELMEnsemble predict")                         
+        # print("ELMEnsemble predict")                         
         check_is_fitted(self, ['X_', 'y_'])
         X_predict = check_array(X_predict)
         
@@ -140,7 +146,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         residuals : numpy array of shape (n_samples, n_estimators)
             Training residuals for all models.
         '''
-        print("ELMEnsemble _collect") 
+        # print("ELMEnsemble _collect") 
         n_obs = self.X_.shape[0]
         n_predict = self.X_predict_.shape[0]
         N = self.n_neurons
@@ -171,7 +177,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Residual sum of squares averaged over all ELMs.
             
         '''
-        print("ELMEnsemble _AvgRSS") 
+        # print("ELMEnsemble _AvgRSS") 
         sq_residuals = np.square(residuals)
         RSS = sq_residuals.sum(axis=0)
         ARSS = RSS.mean()
@@ -194,12 +200,8 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         muTmu : Numpy matrix of shape (n_predict)
             Unbiased estimate of the squared norm of mu
             
-        References
-        ----------
-        Equation (10)
-            
         '''    
-        print("ELMEnsemble _muTmu_estim") 
+        # print("ELMEnsemble _muTmu_estim") 
         N = self.n_estimators
         
         z = np.einsum('pnm, nsm -> psm', H_predict, H_pinvs, optimize = 'greedy')
@@ -225,7 +227,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         ----------
         estimate : string, optional
             Estimate to use, "naive" or "bias-reduced". The "bias-reduced" estimate 
-            is highly recomended, see remark below. Default is "bias-reduced".
+            is highly recommended, see remark below. Default is "bias-reduced".
         
         Returns
         -------
@@ -238,22 +240,25 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         ------
         The "naive" estimate was proposed in a proceeding of the 28th European 
         Symposium on Artificial Neural Networks, Computational Intelligence and 
-        Machine Learning (ESANN 2020). This work motivated the paper "On Extreme 
-        Learning Machine Model Variance" in which the "bias-reduced" estimate was 
+        Machine Learning (ESANN 2020). This work motivated the Neurocomputing 
+        paper (see references below) in which the "bias-reduced" estimate was 
         proposed. In particular, it was shown that the later as a lower bias and, 
         therefore, was recommended. The "naive estimate" is still available for
-        comparison and reprodicibility purposes only.
+        comparison and reproducibility purposes only.
             
         References
-        ---------
-        F. Guignard, F. Amato and M. Kanevski (in prep). On Extreme Learning 
-        Machine Model Variance.        
+        ----------
+        F. Guignard, F. Amato and M. Kanevski. 
+        Uncertainty Quantification in Extreme Learning Machine:  
+        Analytical Developments, Variance Estimates and Confidence Intervals, 
+        Neurocomputing 456 (2021) 436-449.
         
-        F. Guignard, M. Laib and M. Kanevski (in press). Model Variance for 
-        Extreme Learning Machine. 
-            
+        F. Guignard, M. Laib and M. Kanevski. 
+        Model Variance for Extreme Learning Machine,
+        ESANN (2020) 703-708.
+
         ''' 
-        print("ELMEnsemble homoskedastic_variance")                       
+        # print("ELMEnsemble homoskedastic_variance")                       
         n_obs = self.X_.shape[0]
         H_pinvs, H_predict, residuals = self._collect()
         
@@ -284,7 +289,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
              Hidden matrices.
 
         '''
-        print("ELMEnsemble _collect_Hs")  
+        # print("ELMEnsemble _collect_Hs")  
         n_obs = self.X_.shape[0]
         N = self.n_neurons
         M = self.n_estimators
@@ -311,7 +316,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             output weights covariance matrix. 
             
         ''' 
-        print("ELMEnsemble _Sigma_estim")  
+        # print("ELMEnsemble _Sigma_estim")  
         n_obs = corr_res.shape[0]
         rrT = np.einsum('im, jm -> ijm', corr_res, corr_res, optimize = 'greedy')
         Sigma = np.zeros(rrT.shape)
@@ -337,7 +342,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             S1 estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _S1_estim")  
+        # print("ELMEnsemble _S1_estim")  
         S1 = np.einsum('pim, ijm, pjm -> p', z, Sigma, z, optimize = 'greedy')
         S1 = S1 /self.n_estimators
         
@@ -358,7 +363,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             nu estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _nu_estim") 
+        # print("ELMEnsemble _nu_estim") 
         nu = np.einsum('pim, ijm -> pj', z, Sigma, optimize = 'greedy')
         nu = nu / self.n_estimators
         return nu
@@ -380,7 +385,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             S2 estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _S2_estim") 
+        # print("ELMEnsemble _S2_estim") 
         nuTmu = np.einsum('ps, ps -> p', mu, nu, optimize = 'greedy')
         S2 = self.n_estimators * nuTmu - S1
         S2 = S2 / (self.n_estimators - 1)
@@ -402,7 +407,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Averaged quadratic form in z, see section 3.3.
             
         '''    
-        print("ELMEnsemble _V_estim") 
+        # print("ELMEnsemble _V_estim") 
         V = np.einsum('pim, ij, pjm -> p', z, U, z, optimize = 'greedy')
         V = V /self.n_estimators
         
@@ -427,7 +432,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             S3 estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _S3_estim") 
+        # print("ELMEnsemble _S3_estim") 
         muTUmu = np.einsum('pi, ij, pj -> p', mu, U, mu, optimize = 'greedy')
         S3 = self.n_estimators**2 * muTUmu - self.n_estimators*V - 2*(self.n_estimators-1)*S2
         S3 = S3 / ((self.n_estimators - 1) * (self.n_estimators - 2))
@@ -449,7 +454,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Approximation of S1 estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _S1_approx") 
+        # print("ELMEnsemble _S1_approx") 
         S1_app = np.einsum('psm, sm, psm -> p', z, Omega, z, optimize = 'greedy')
         S1_app = S1_app/self.n_estimators
         
@@ -470,7 +475,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             approximation of nu estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _nu_approx") 
+        # print("ELMEnsemble _nu_approx") 
         nu_app = np.einsum('psm, sm -> ps', z, Omega, optimize = 'greedy')
         nu_app = nu_app / self.n_estimators
         return nu_app
@@ -490,7 +495,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Averaged quadratic form in z, see section 3.3.
             
         '''    
-        print("ELMEnsemble _V_approx") 
+        # print("ELMEnsemble _V_approx") 
         V_app = np.einsum('psm, s, psm -> p', z, U_app, z, optimize = 'greedy')
         V_app = V_app /self.n_estimators
         
@@ -515,7 +520,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
             Approximation of S3 estimate, see section 3.3.
             
         '''    
-        print("ELMEnsemble _S3_approx") 
+        # print("ELMEnsemble _S3_approx") 
         M = self.n_estimators
         muTUmu = np.einsum('ps, s, ps -> p', mu, U_app, mu, optimize = 'greedy')
         S3_app = M**2 * muTUmu - M*V_app - 2*(M-1)*S2_app
@@ -532,9 +537,9 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         ----------
         estimate : string, optional
             Estimate to use, "naive", "S1, S2" or "S3". The "S3" estimate 
-            is highly recomended, see remark below. Default is "S3".
+            is highly recommended, see remark below. Default is "S3".
         approx : bool, optional
-            Speed up computation at the expense of a sligntly approximate variance 
+            Speed up computation at the expense of a slightly approximate variance 
             estimation. Default is 'False'.
         corrected_residuals : bool, optional
             If True, the corrected residuals used to compute HC3 are squared and
@@ -552,27 +557,29 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         ------
         The "naive" estimate was proposed in a proceeding of the 28th European 
         Symposium on Artificial Neural Networks, Computational Intelligence and 
-        Machine Learning (ESANN 2020). This work motivated the paper "On Extreme 
-        Learning Machine Model Variance" in which the "S1", "S2" and "S3" estimates 
-        were proposed. In particular, it was shown empirically that "S3" as a lower 
-        bias among others and, therefore, was recommended. The "naive", "S1" and 
-        "S2" estimates are still available for comparison and reprodicibility 
-        purposes only.
+        Machine Learning (ESANN 2020). This work motivated the Neurocomputing 
+        paper (see references below) in which the "bias-reduced" estimate was 
+        proposed. In particular, it was shown that the later as a lower bias and, 
+        therefore, was recommended. The "naive estimate" is still available for
+        comparison and reproducibility purposes only.
             
         References
         ----------
-        F. Guignard, F. Amato and M. Kanevski (in prep). On Extreme Learning 
-        Machine Model Variance.
-
-        F. Guignard, M. Laib and M. Kanevski (in press). Model Variance for 
-        Extreme Learning Machine. 
+        F. Guignard, F. Amato and M. Kanevski. 
+        Uncertainty Quantification in Extreme Learning Machine:  
+        Analytical Developments, Variance Estimates and Confidence Intervals, 
+        Neurocomputing 456 (2021) 436-449.
         
-        J.G. MacKinnon, H. White (1985). Some heteroskedasticity-consistent 
+        F. Guignard, M. Laib and M. Kanevski. 
+        Model Variance for Extreme Learning Machine,
+        ESANN (2020) 703-708.
+        
+        J.G. MacKinnon, H. White. Some heteroskedasticity-consistent 
         covariance matrix estimators with improved finite sample properties, 
-        Journal of econometrics, 29, 3, 305--325. 
+        Journal of econometrics 29 (3) (1985) 305--325. 
                     
         ''' 
-        print("ELMEnsemble heteroskedastic_variance")                                     
+        # print("ELMEnsemble heteroskedastic_variance")                                     
         H_pinvs, H_predict, residuals = self._collect()
         Hs = self._collect_Hs()
         
@@ -588,7 +595,7 @@ class ELMEnsemble(BaseEstimator, RegressorMixin):
         else :
             raise TypeError("Only booleans are allowed for the 'corrected_residuals' argument")
         
-        if estimate == 'naive' :   # naive estimation from ESANN proceding
+        if estimate == 'naive' :   # naive estimation from ESANN proceeding
             mu = z.mean(axis=2)
             if approx == False :
                 Sigma = self._Sigma_estim(corr_res)
@@ -669,7 +676,7 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
         None.
 
         '''
-        print("ELMEnsembleRidge __init__")      
+        # print("ELMEnsembleRidge __init__")      
         ELMEnsemble.__init__(self, n_estimators, n_neurons, activation, 
                             weight_distr, weight_scl, n_jobs, random_state)  
         self.alpha = alpha             
@@ -677,7 +684,7 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
                 
     def fit(self, X, y):
         '''
-        Training for ELM ensemble.
+        Training for regularized ELM ensemble.
         Initialize random hidden weights and compute output weights 
         for all ELM models.
         
@@ -692,8 +699,14 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
         -------
         Self.
         
+        Reference
+        ---------
+        W. Deng, Q. Zheng, L. Chen,  
+        Regularized extreme learning machine, 
+        IEEE symposium on computational intelligence and data mining, 2009, pp. 389–395.
+        
         '''    
-        print("ELMEnsembleRidge fit")                
+        # print("ELMEnsembleRidge fit")                
         self.check = X.shape
         X, y = check_X_y(X, y)
         
@@ -736,7 +749,7 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
         residuals : numpy array of shape (n_samples, n_estimators)
             Training residuals for all models.
         '''
-        print("ELMEnsembleRidge _collect") 
+        # print("ELMEnsembleRidge _collect") 
         n_obs = self.X_.shape[0]
         n_predict = self.X_predict_.shape[0]
         N = self.n_neurons
@@ -754,7 +767,7 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
             
         return H_alphas, H_predict, residuals
 
-    def homoskedastic_variance(self, estimate='bias-reduced', approx=False):
+    def homoskedastic_variance(self, estimate='bias-reduced'):
         '''
         Compute an homoskedastic variance estimation of the model at last predicted points.
             
@@ -762,10 +775,7 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
         ----------
         estimate : string, optional
             Estimate to use, 'naive' or 'bias-reduced'. The 'bias-reduced' estimate 
-            is recomended, see remark below. Default is 'bias-reduced'.
-        approx : bool, optional
-            Speed up computation using the Hastie and Tibshirani approximation
-            of effective degrees of freedom. Default is 'False'.
+            is recommended, see remark below. Default is 'bias-reduced'.
 
         Returns
         -------
@@ -774,47 +784,39 @@ class ELMEnsembleRidge(ELMEnsemble, BaseEstimator, RegressorMixin):
         noise_estimate : numpy.float
             Noise estimate.
  
-        Remark
+        Remarks
         ------
         The "naive" estimate was proposed in a proceeding of the 28th European 
         Symposium on Artificial Neural Networks, Computational Intelligence and 
-        Machine Learning (ESANN 2020). This work motivate the paper "On Extreme 
-        Learning Machine Model Variance" in which the "bias-reduced" estimate was 
+        Machine Learning (ESANN 2020). This work motivated the Neurocomputing 
+        paper (see references below) in which the "bias-reduced" estimate was 
         proposed. In particular, it was shown that the later as a lower bias and, 
         therefore, was recommended. The "naive estimate" is still available for
-        comparison and reprodicibility purposes only.
-                       
+        comparison and reproducibility purposes only.
+            
         References
         ----------
-        F. Guignard, F. Amato and M. Kanevski (in prep). On Extreme Learning 
-        Machine Model Variance.
+        F. Guignard, F. Amato and M. Kanevski. 
+        Uncertainty Quantification in Extreme Learning Machine:  
+        Analytical Developments, Variance Estimates and Confidence Intervals, 
+        Neurocomputing 456 (2021) 436-449.
         
-        F. Guignard, M. Laib and M. Kanevski (in press). Model Variance for 
-        Extreme Learning Machine. 
-        
-        T.J. Hastie, R.J. Tibshirani (1990). 
-        Generalized additive models, CRC press. 
+        F. Guignard, M. Laib and M. Kanevski. 
+        Model Variance for Extreme Learning Machine,
+        ESANN (2020) 703-708.
             
         ''' 
-        print("ELMEnsembleRidge homoskedastic_variance")                 
+        # print("ELMEnsembleRidge homoskedastic_variance")                 
         n_obs = self.X_.shape[0]
         H_alphas, H_predict, residuals = self._collect()
         Hs = self._collect_Hs()
-
-        # Compute the effective degrees of freedom  
-        Ptrace = np.einsum('ijm, jim -> m', Hs, H_alphas, optimize = 'greedy') # Traces of the Hat matrices 
-        if approx == False :
-            PPtrace = np.einsum('ijm, jkm, klm, lim -> m', Hs, H_alphas, Hs, H_alphas, 
-                                optimize = 'greedy')  # Traces of the squared Hat matrices
-            gamma = 2*Ptrace.mean() - PPtrace.mean()
-            
-        elif approx == True :
-            gamma = -1/2 + 5/4*Ptrace.mean()        
-            
-        else:
-            raise TypeError("Only booleans are allowed for the 'approx' argument")
+        Hs = Hs.transpose(2,0,1)
+        eigenHTHs = np.square(np.linalg.svd(Hs, full_matrices= False, compute_uv=False, hermitian=False))
+        eigenP = (eigenHTHs/(eigenHTHs + self.alpha))
+        eigenP2 = np.square(eigenP)
+        self.gamma_ = (2*eigenP - eigenP2).sum() / self.n_estimators
         
-        ddof = n_obs - gamma 
+        ddof = n_obs - self.gamma_ 
                 
         # Compute noise estimation
         ARSS = self._AvgRSS(residuals)
@@ -869,7 +871,7 @@ class ELMEnsembleRidgeCV(ELMEnsembleRidge, BaseEstimator, RegressorMixin):
         None.
 
         '''        
-        print("ELMEnsembleRidgeCV __init__")
+        # print("ELMEnsembleRidgeCV __init__")
         ELMEnsemble.__init__(self, n_estimators, n_neurons, activation, 
                              weight_distr, weight_scl, n_jobs, random_state)  
         self.alphas = alphas         
@@ -877,7 +879,7 @@ class ELMEnsembleRidgeCV(ELMEnsembleRidge, BaseEstimator, RegressorMixin):
     
     def fit(self, X, y):
         '''
-        Training for ELM ensemble.
+        Training for regularized ELM ensemble, with Generalized Cross Validation.
         Initialize random hidden weights and compute output weights 
         for all ELM models.
         
@@ -892,8 +894,18 @@ class ELMEnsembleRidgeCV(ELMEnsembleRidge, BaseEstimator, RegressorMixin):
         -------
         Self.
         
+        References
+        ----------
+        W. Deng, Q. Zheng, L. Chen,  
+        Regularized extreme learning machine, 
+        IEEE symposium on computational intelligence and data mining, 2009, pp. 389–395.
+        
+        G. H. Golub, M. Heath, G. Wahba, 
+        Generalized cross-validation as a method for choosing a good ridge parameter,
+        Technometrics 21 (2) (1979) 215–223.
+        
         '''    
-        print("ELMEnsembleRidgeCV fit")        
+        # print("ELMEnsembleRidgeCV fit")        
         self.check = X.shape
         X, y = check_X_y(X, y)
         
@@ -922,5 +934,103 @@ class ELMEnsembleRidgeCV(ELMEnsembleRidge, BaseEstimator, RegressorMixin):
             self.estimators_[i] = elm
 
         return self    
+    
+    def _collect_alphas_opt(self):
+        '''
+        Collect alpha_opt in each model. Complete the _collect() function for the
+        homoskedastic GCV case.
 
+        Returns
+        -------
+        alphas_opt : numpy array of shape (n_estimators)
+             alpha_opt for each model of the ensemble.
 
+        '''
+        # print("ELMEnsembleRidgeCV _collect_alphas_opt")  
+        M = self.n_estimators
+        
+        alphas_opt = np.zeros((M))  
+        for i in range(M):   
+            elm = self.estimators_[i]
+            alphas_opt[i] = elm.alpha_opt  
+            
+        return alphas_opt
+
+    def homoskedastic_variance(self, estimate='bias-reduced'):
+        '''
+        Compute an homoskedastic variance estimation of the model at last predicted points.
+            
+        Parameters
+        ----------
+        estimate : string, optional
+            Estimate to use, 'naive' or 'bias-reduced'. The 'bias-reduced' estimate 
+            is recommended, see remark below. Default is 'bias-reduced'.
+
+        Returns
+        -------
+        var_predict : numpy array of shape (n_samples)
+            Model variance estimation for y_predict.
+        noise_estimate : numpy.float
+            Noise estimate.
+ 
+        Remarks
+        ------
+        The "naive" estimate was proposed in a proceeding of the 28th European 
+        Symposium on Artificial Neural Networks, Computational Intelligence and 
+        Machine Learning (ESANN 2020). This work motivated the Neurocomputing 
+        paper (see references below) in which the "bias-reduced" estimate was 
+        proposed. In particular, it was shown that the later as a lower bias and, 
+        therefore, was recommended. The "naive estimate" is still available for
+        comparison and reproducibility purposes only.
+            
+        References
+        ----------
+        F. Guignard, F. Amato and M. Kanevski. 
+        Uncertainty Quantification in Extreme Learning Machine:  
+        Analytical Developments, Variance Estimates and Confidence Intervals, 
+        Neurocomputing 456 (2021) 436-449.
+        
+        F. Guignard, M. Laib and M. Kanevski. 
+        Model Variance for Extreme Learning Machine,
+        ESANN (2020) 703-708.
+            
+        ''' 
+        # print("ELMEnsembleRidgeCV homoskedastic_variance")                 
+        n_obs = self.X_.shape[0]
+        H_alphas, H_predict, residuals = self._collect()
+        Hs = self._collect_Hs()          
+        Hs = Hs.transpose(2,0,1)
+        alphas_opt = self._collect_alphas_opt()                                 ### change from non GCV case
+        alphas_opt = alphas_opt.reshape((alphas_opt.shape[0], 1))               ### change from non GCV case
+        alphas_opt = alphas_opt.repeat(self.n_neurons, axis = 1)                ### change from non GCV case
+        eigenHTHs = np.square(np.linalg.svd(Hs, full_matrices= False, compute_uv=False, hermitian=False))
+        eigenP = (eigenHTHs/(eigenHTHs + alphas_opt))                           ### change from non GCV case
+        eigenP2 = np.square(eigenP)
+        self.gamma_ = (2*eigenP - eigenP2).sum() / self.n_estimators
+        
+        ddof = n_obs - self.gamma_ 
+                
+        # Compute noise estimation
+        ARSS = self._AvgRSS(residuals)
+        noise = ARSS / ddof
+        
+        #Compute model variance induced by noise
+        muTmu = self._muTmu_estim(H_alphas, H_predict, estimate)
+        var1 = noise * muTmu
+        
+        # Compute model variance induced by the random weights
+        var2 = self.y_var_/self.n_estimators
+
+        var_predict = var1 + var2 
+            
+        return var_predict, noise 
+
+# Future improvement
+# This homoskedastic_variance function could be re-written here : 
+# 1) Indeed, the svd is done twice when this class is used, because the homoskedastic_variance 
+#    function is inherited from the class ELMEnsembleRidge, where the svd is not done 
+#    during the fitting process. It could be done one time and store during the 
+#    fitting process of the ELMEnsembleRidgeCV and the ELMRidgeCV classes (elm_estimator.py).
+# 2) The writing could be optimize, as there is not too much changes comparing to 
+#    the ELMEnsembleRidge corresponding function.        
+# 3) Eventually put the alphas_opt in atribute during the fitting process.
